-- Coding standards brought to you by the East India Trading Company in association with the Save the Camel Foundation
local L = LibStub("AceLocale-3.0"):GetLocale("BleedingHearts")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local SM = LibStub("LibSharedMedia-3.0")
local playerName, playerGUID, playerClass, db

assert(AceGUIWidgetLSMlists, "Utopia requires AceGUI-3.0-SharedMediaWidgets")

local UnitInVehicle	= _G.UnitInVehicle
local UnitInParty	= _G.UnitInParty
local UnitInRaid	= _G.UnitInRaid
local UnitGUID		= _G.UnitGUID
local GetSpellInfo	= _G.GetSpellInfo
local next, pairs, select, tonumber, strmatch, type = _G.next, _G.pairs, _G.select, _G.tonumber, _G.strmatch, _G.type
local floor, ceil, abs, tremove, tinsert = _G.floor, _G.ceil, _G.abs, _G.tremove, _G.tinsert

--[===[@debug@
local d
do
	BHDEBUG = true
	local z
	local function dd(str)
		local r, g, b = z%3, z%4, z%5
		z = z + 1
		return format("|cFF%02X%02X%02X%s|r", r / 4 * 255, g / 4 * 255, b / 4 * 255, str)
	end
	function d(fmt, ...)
		if (BHDEBUG) then
			z = 3
			fmt = fmt:gsub("(%%[sdqxf])", dd)
			ChatFrame1:AddMessage("|cFFFF8080BH:|r "..format(fmt, ...), 0.8, 0.8, 0.8)
		end
	end
end
--@end-debug@]===]

local new, del, copy, deepDel
do
	local list = setmetatable({},{__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			t = {...}
			return t
		end
	end
	function del(t)
		if (t) then
			wipe(t)
			t[''] = true
			t[''] = nil
			list[t] = true
		end
	end
	function deepDel(t)
		if (t) then
			for k,v in pairs(t) do
				if type(v) == "table" then
					deepDel(v)
				end
				t[k] = nil
			end
			t[''] = true
			t[''] = nil
			list[t] = true
		end
	end
end

local module = LibStub("AceAddon-3.0"):NewAddon("BleedingHearts", "AceConsole-3.0", "AceEvent-3.0")
--[===[@debug@
BH = module
--@end-debug@]===]
module.version = tonumber(strmatch("$Revision: 21 $", "(%d+)")) or 0
module.version = max(module.version, tonumber(strmatch("21", "(%d+)")) or 0)

local LDB = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("BleedingHearts", {
	type = "launcher",
	label = L["TITLE"],
	icon = select(3, GetSpellInfo(48672)),
})

-- OnClick
function LDB.OnClick(self, button)
	if (not IsModifierKeyDown()) then
		if (button == "RightButton") then
			module:OpenConfig()
		elseif (button == "LeftButton") then
			if (module.frame and module.frame:IsShown()) then
				module:HideFrame()
			else
				module:ShowFrame(1)
			end
		end
	end
end

-- OnTooltipShow
function LDB.OnTooltipShow(tt)
	tt:AddLine(module.title)
end

local flatskinhidden = function() return db.skin ~= "flat" end

module.options = {
	type = 'group',
	order = 15,
	name = L["TITLE"],
	desc = L["Configuration"],
	handler = module,
	get = function(info) return db[info[#info]] end,
	set = function(info, value) db[info[#info]] = value module:ShowWhen() end,
	args = {
		General = {
			type = 'group',
			order = 1,
			name = L["General Settings"],
			desc = L["General Settings"],
			args = {
				ticks = {
					type = 'toggle',
					name = L["Show Ticks"].." (NYI)",
					desc = L["Show last ticks for dots on current target"],
					order = 5,
				},
				show = {
					type = 'group',
					name = L["Show When"],
					desc = L["When to display"],
					guiInline = true,
					order = 100,
					get = function(info) return db.show[info[#info]] end,
					set = function(info, value)
						db.show[info[#info]] = value
						module:ShowWhen()
					end,
					args = {
						desc = {
							order = 0,
							type = "description",
							name = L["Indicate when to show the display."],
						},
  						always = {
							type = 'toggle',
							name = L["Always"],
							desc = L["Always show"],
							order = 10,
						},
						solo = {
							type = 'toggle',
							name = SOLO,
							desc = L["Show when solo"],
							disabled = function() return db.show.always end,
							order = 20,
						},
						party = {
							type = 'toggle',
							name = PARTY,
							desc = L["Show when in a party"],
							disabled = function() return db.show.always end,
							order = 30,
						},
						raid = {
							type = 'toggle',
							name = RAID,
							desc = L["Show when in a raid"],
							disabled = function() return db.show.always end,
							order = 40,
						},
					},
				},
				incombat = {
					type = 'group',
					name = L["In-Combat Options"],
					desc = L["In-Combat Options"],
					guiInline = true,
					order = 200,
					args = {
						desc = {
							order = 1,
							type = "description",
							name = L["In-combat specific options."],
						},
						inalpha = {
							type = "range",
							name = L["Alpha"],
							desc = L["Alpha level in-combat"],
							order = 10,
							min = 0,
							max = 1,
							step = 0.01,
						},
					},
				},
				oocombat = {
					type = 'group',
					name = L["Out-of-Combat Options"],
					desc = L["Out-of-Combat Options"],
					guiInline = true,
					order = 300,
					args = {
						desc = {
							order = 1,
							type = "description",
							name = L["Out-of-combat specific options."],
						},
						outalpha = {
							type = "range",
							name = L["Alpha"],
							desc = L["Alpha level out-of-combat"],
							order = 10,
							min = 0,
							max = 1,
							step = 0.01,
						},
					},
				},
				icon = {
					type = 'group',
					name = L["Icon Options"],
					desc = L["Icon Options"],
					guiInline = true,
					order = 400,
					args = {
						showCooldown = {
							type = "toggle",
							order = 10,
							name = L["Show Cooldown"],
							desc = L["Show cooldown indication on icons"],
						},
						drawEdge = {
							type = "toggle",
							order = 15,
							name = L["Edge Highlight"],
							desc = L["Show leading edge on cooldown for clarity"],
						},
						showCountdown = {
							type = "toggle",
							order = 20,
							name = L["Show Countdown"],
							desc = L["Show buff countdown in icons"],
						},
						minCountdown = {
							type = "range",
							name = L["Minimum Duration"],
							desc = L["Set minumum time left before the countdown shows"],
							order = 2,
							min = 5,
							max = 90,
							step = 1,
							bigStep = 10,
							order = 30,
						},
					},
				},
				display = {
					type = 'group',
					name = L["Display Options"],
					desc = L["Display Options"],
					guiInline = true,
					order = 500,
					args = {
						desc = {
							order = 0,
							type = "description",
							name = L["The scale and location of the display can be changed by holding down shift and using mousewheel or dragging."],
						},
						skin = {
							order = 10,
							type = "select",
							name = L["Skin"],
							desc = L["Select the skin to use"],
							values = {flat = L["Flat"], thermometer = L["Thermometer"]},
							set = function(info, v)
								db.skin = v
								module.callbacks:Fire("SkinChanged")
							end,
						},
						flattexture = {
							type = 'select', dialogControl = 'LSM30_Statusbar',
							name = L["Flat Skin Texture"],
							desc = L["Set the texture for the bar"],
							hidden = flatskinhidden,
							values = AceGUIWidgetLSMlists.statusbar,
							order = 20,
							set = function(info, v)
								db.flattexture = v
								module.callbacks:Fire("SkinChanged")
							end,
						},
						flatwidth = {
							type = "range",
							name = L["Width"],
							desc = L["Width of flat skin"],
							hidden = flatskinhidden,
							set = function(info, v)
								db.flatwidth = v
								module.callbacks:Fire("SkinChanged")
							end,
							order = 30,
							min = 100,
							max = 500,
							step = 1,
						},
						scale = {
							type = "range",
							name = L["Scale"],
							desc = L["Scale"],
							order = 100,
							min = 0.2,
							max = 5,
							step = 0.01,
							bigStep = 0.1,
						}
					},
				},
			},
		},
	},
}
AceConfig:RegisterOptionsTable("BleedingHearts", module.options)

module.monitor = {
	[GetSpellInfo(48672)] = "ROGUE",					-- Rupture
	[GetSpellInfo(1822)] = "DRUID",						-- Rake
	[GetSpellInfo(55095)] = "DEATHKNIGHT",				-- Frost Fever
	[GetSpellInfo(55078)] = "DEATHKNIGHT",				-- Blood Plague
}

do
	local function barOnUpdate(self, elapsed)
		if (self.min and self.targetMin) then
			local diff = self.min - self.targetMin
			self.min = self.min - diff * (elapsed * 10)
			if (abs(self.min - self.targetMin) < 1) then
				self.min = self.targetMin
				self.targetMin = nil
			end
		end
		if (self.max and self.targetMax) then
			local diff = self.targetMax - self.max
			self.max = self.max + diff * (elapsed * 10)
			if (abs(self.max - self.targetMax) < 1) then
				self.max = self.targetMax
				self.targetMax = nil
			end
		end
		if (not self.pos) then
			self.pos = self.target
			self.target = nil
		elseif (self.pos and self.target) then
			local diff = (self.pos - self.target)
			self.pos = self.pos - diff * (elapsed * 10)
			if (abs(self.pos - self.target) < 1 or (self.pos > self.target + 100000)) then
				self.pos = self.target
				self.target = nil
			end
		end
		self:Draw()
		if (not self.target and not self.targetMin and not self.targetMax) then
			self:SetScript("OnUpdate", nil)
		end
	end

	-- bar.Draw
	local function barDraw(self)
		self.textMin:SetFormattedText("%d", self.min or 0)
		self.textMax:SetFormattedText("%d", self.max or 0)

		self.bar:SetMinMaxValues(self.min, self.max)
		self.bar:SetValue(self.pos or self.min or 0)

		self:SetTextValue()
	end

	-- bar.SetValue
	local function barSetValue(self, val)
		if (not self.min) then
			self.min = floor(val / 100) * 100
			self.max = ceil(val / 100) * 100
		else
			if (val < self.min) then
				self.targetMin = floor(val / 100) * 100
			end
			if (val > self.max) then
				self.targetMax = ceil(val / 100) * 100
			end
		end
		self.target = val
		self:SetScript("OnUpdate", barOnUpdate)
	end

	-- bar.Refresh
	local function barRefresh(self)
		local unitInfo = module.active and module.active[UnitGUID("target")]
		if (unitInfo) then
			for spellName,info in pairs(unitInfo) do
				info.auraEndTime = nil				-- Forces a refresh of buff times
			end
		end

		self:SetScript("OnUpdate", barOnUpdate)
	end

	-- bar.SetSkin
	local function barSetSkin(self)
		local skin = db.skin
		self:ClearAllPoints()
		if (skin == "thermometer") then
			self:GetParent():SetWidth(48)
			self:GetParent():SetHeight(196)
			self:SetAllPoints()

			if (not self.back) then
				self.back = self:CreateTexture(nil, "BACKGROUND")
			else
				self.back:ClearAllPoints()
			end
			self.back:SetTexture("Interface\\AddOns\\BleedingHearts\\Textures\\Therm-Back")
			self.back:SetAllPoints()

			if (not self.bulb) then
				self.bulb = self:CreateTexture(nil, "BORDER")
			else
				self.bulb:Show()
				self.bulb:ClearAllPoints()
			end
			self.bulb:SetTexture("Interface\\AddOns\\BleedingHearts\\Textures\\Therm-Bulb")
			self.bulb:SetWidth(module.frame:GetWidth())
			self.bulb:SetHeight(44)
			self.bulb:SetPoint("BOTTOM")
			self.bulb:SetTexCoord(0, 1, 0.13, 0.985)

			if (not self.bar) then
				self.bar = CreateFrame("StatusBar", nil, self)
			else
				self.bar:ClearAllPoints()
			end
			self.bar:SetPoint("BOTTOMLEFT", self.bulb, "TOPLEFT")
			self.bar:SetPoint("TOPRIGHT", 0, -12)
			self.bar:SetOrientation("VERTICAL")
			self.bar:SetStatusBarTexture("Interface\\AddOns\\BleedingHearts\\Textures\\Therm-Bar")
			self.bar:SetStatusBarColor(1, 1, 1)

			if (not self.shine) then
				self.shine = self.bar:CreateTexture(nil, "OVERLAY")
			else
				self.shine:Show()
				self.shine:ClearAllPoints()
			end
			self.shine:SetTexture("Interface\\AddOns\\BleedingHearts\\Textures\\Therm-Shine")
			self.shine:SetPoint("TOPLEFT", 0, 12)
			self.shine:SetPoint("BOTTOMRIGHT", 0, -44)

			if (not self.textMin) then
				self.textMin = self.bar:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			else
				self.textMin:ClearAllPoints()
			end
			self.textMin:SetJustifyH("LEFT")
			self.textMin:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 28, 40)
			self.textMin:SetTextColor(1, 0, 0)

			if (not self.textMax) then
				self.textMax = self.bar:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			else
				self.textMax:ClearAllPoints()
			end
			self.textMax:SetJustifyH("LEFT")
			self.textMax:SetPoint("TOPLEFT", self, "TOPLEFT", 28, -6)
			self.textMax:SetTextColor(1, 1, 1)

			if (not self.text) then
				self.text = self.bar:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			end
			self.text:SetJustifyH("CENTER")
			self.text:SetTextColor(0, 1, 0)

			function self:SetTextValue()
				if (self.pos) then
					local range = self.max - self.min
					local y = self.bar:GetHeight() / range * (self.pos - self.min)
					self.text:ClearAllPoints()
					self.text:SetPoint("LEFT", self.bar, "BOTTOMRIGHT", -20, y)
					self.text:SetFormattedText("%d", self.pos)
				end
			end
		else
			if (self.bulb) then
				self.bulb:Hide()
			end
			if (self.shine) then
				self.shine:Hide()
			end
		end

		if (skin == "flat") then
			self:GetParent():SetWidth(db.flatwidth or 250)
			if (playerClass == "DEATHKNIGHT") then
				self:GetParent():SetHeight(40)
			else
				self:GetParent():SetHeight(20)
			end
			self:SetAllPoints()

			if (not self.back) then
				self.back = self:CreateTexture(nil, "BACKGROUND")
			else
				self.back:ClearAllPoints()
			end
			self.back:SetTexture(0, 0, 0)
			self.back:SetAllPoints()

			if (not self.bar) then
				self.bar = CreateFrame("StatusBar", nil, self)
			else
				self.bar:ClearAllPoints()
			end
			self.bar:SetPoint("BOTTOMLEFT")
			self.bar:SetPoint("TOPRIGHT")
			self.bar:SetOrientation("HORIZONTAL")
			self.bar:SetStatusBarTexture(SM:Fetch("statusbar", db.flattexture))
			self.bar:SetStatusBarColor(0.8, 0.2, 0.2)

			if (not self.textMin) then
				self.textMin = self.bar:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			else
				self.textMin:ClearAllPoints()
			end
			self.textMin:SetJustifyH("LEFT")
			self.textMin:SetPoint("TOPLEFT")
			self.textMin:SetTextColor(1, 0, 0)

			if (not self.textMax) then
				self.textMax = self.bar:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			else
				self.textMax:ClearAllPoints()
			end
			self.textMax:SetJustifyH("RIGHT")
			self.textMax:SetPoint("TOPRIGHT")
			self.textMax:SetTextColor(1, 1, 1)

			if (not self.text) then
				self.text = self.bar:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			end
			self.text:SetJustifyH("CENTER")
			self.text:SetTextColor(0, 1, 0)

			function self:SetTextValue()
				if (self.pos) then
					local range = self.max - self.min
					local x = self.bar:GetWidth() / range * (self.pos - self.min)
					self.text:ClearAllPoints()
					self.text:SetPoint("BOTTOM", self.bar, "TOPLEFT", x, 0)
					self.text:SetFormattedText("%d", self.pos)
					if (self.text:GetLeft() < self.bar:GetLeft()) then
						self.text:ClearAllPoints()
						self.text:SetPoint("BOTTOMLEFT", self.bar, "TOPLEFT")
					elseif (self.text:GetRight() > self.bar:GetRight()) then
						self.text:ClearAllPoints()
						self.text:SetPoint("BOTTOMRIGHT", self.bar, "TOPRIGHT")
					end
				end
			end
		end

		self:SetTextValue()
	end

	-- CreateBar
	function module:CreateBar(mode)
		local bar = CreateFrame("Frame", nil, module.frame)
		bar.mode = mode

		--bar:SetAllPoints()

		bar.SetValue		= barSetValue
		bar.Draw			= barDraw
		bar.Refresh			= barRefresh
		bar.SetSkin			= barSetSkin

		self:RegisterCallback("SkinChanged", barSetSkin, bar)

		return bar
	end
end

-- RuptureCalcs
local base = {10, 15, 22, 32, 44, 68, 81, 122, 128}
local baseAdd = {2, 3, 4, 5, 7, 8, 9, 15, 18}
local function RuptureCalcs(rank, ap, points)
	-- The calc's are not quite right, but it's close to what was on WoWhead
	-- Returns approximate tick damage
	local full = ((base[rank] + (baseAdd[rank] * points)) + (0.015 + 0.009 * points) * ap) * (points + 3)
	return full / (6 + (points * 2))
end

do
	local pool = {}
	-- apbar.CreateIcon
	local function apbarCreateIcon(self)
		local icon = CreateFrame("Frame", nil, self)
		icon:SetWidth(20)
		icon:SetHeight(20)
		icon.tex = icon:CreateTexture(nil, "OVERLAY")
		icon.tex:SetAllPoints()
		return icon
	end

	-- apbar.AquireIcon
	local function apbarAquireIcon(self)
		local icon = tremove(pool) or self:CreateIcon()
		icon:Show()
		return icon
	end

	-- apbar.FreeIcon
	local function apbarFreeIcon(self, icon)
		icon:Hide()
		tinsert(pool, icon)
	end
	
	-- apbar.event.ChangedAP
	local function apbarEventChangedAP(self, e, newAP)
		self:SetValue(newAP)
	end

	-- apbar.event.AuraApplied
	local function apbarEventAuraApplied(self, e, guid, spellName, spellId)
		if (guid == UnitGUID("target")) then
			self:Refresh()
		end
	end

	-- apbar.event.AuraRemoved
	local function apbarEventAuraRemoved(self, e, guid, spellName, spellId)
		if (guid == UnitGUID("target")) then
			self:Refresh()
		end
	end

	-- countdownOnUpdate
	local function countdownOnUpdate(self, elapsed)
		self.timeLeft = self.timeLeft - elapsed
		if (self.timeLeft < 0 or self.timeLeft > db.minCountdown) then
			self.countdown:Hide()
		else
			self.countdown:Show()
			self.countdown:SetFormattedText(self.timeLeft < 1 and "%.1f" or "%d", self.timeLeft)
		end
	end

	-- apbar.Draw
	local function apbarDraw(self)
		self:SuperDraw()

		local range = self.max - self.min
		local horz = self.bar:GetOrientation() == "HORIZONTAL"
		local height = self.bar:GetHeight()
		local width = self.bar:GetWidth()

		local oldicons = new()
		for spellName,icon in pairs(self.icons) do
			oldicons[spellName] = true
		end

		local unitInfo = module.active[UnitGUID("target")]
		if (unitInfo) then
			local list = new()

			for spellName,info in pairs(unitInfo) do
				tinsert(list, spellName)
				oldicons[spellName] = nil
			end
			sort(list)

			for i,spellName in ipairs(list) do
				local info = unitInfo[spellName]
				local icon = self.icons[spellName]
				if (not icon) then
					icon = self:AquireIcon()
					self.icons[spellName] = icon

					local name, rank, texture = GetSpellInfo(info.id)
					icon.tex:SetTexture(texture)
				end
				icon.ap = info.ap

				--local diff
				--if (info.combos and info.value) then
				--	-- Special case for Rupture, which varies depending on combo points spent
				--	local combos = GetComboPoints("player")
				--	local potential = RuptureCalcs(self.lastRuptureRank, module.currentAP, combos)
				--	diff = potential - info.value
				--else
				--	diff = module.currentAP - info.ap
				--end

				local diff = module.currentAP - info.ap
				if (diff > 0 and (not info.combos or GetComboPoints("player") >= info.combos)) then
					local perc = range / 3 - diff
					local g = min(1, max(0, 2 * perc))

					-- Positive number here means debuff's AP was less than our current, diff shows how much
					icon.tex:SetVertexColor(1, g, 0)
				else
					icon.tex:SetVertexColor(1, 1, 1)
				end

				local x, y
				if (horz) then
					x = width / range * (icon.ap - self.min)
					y = 0
				else
					y = height / range * (icon.ap - self.min)
					x = 0
				end

				for j = 1,i - 1 do
					local icon2 = self.icons[list[j]]
					if (horz) then
						if (icon2.y == 0 and abs(icon2.x - x) <= 10) then
							y = 20
							break
						end
					else
						if (icon2.x == 0 and abs(icon2.y - y) <= 10) then
							x = -20
							break
						end
					end
				end

				local c = icon.count
				if (info.combos) then
					if (not c) then
						c = icon:CreateFontString(nil, "OVERLAY", "GameFontNormalHuge")
						c:SetAllPoints()
						icon.count = c
						c:SetTextColor(1, 1, 1)
					else
						c:Show()
					end
					c:SetFormattedText("%d", info.combos)
				elseif (c) then
					c:Hide()
				end

				icon.y = y
				icon.x = x
				if (horz) then
					icon:SetPoint("BOTTOM", self.bar, "BOTTOMLEFT", x, y)
				else
					icon:SetPoint("LEFT", self.bar, "BOTTOMLEFT", x, y)
				end

				if (not info.auraEndTime) then
					local name, rank, tex, count, debuffType, duration, endTime, caster, isStealable = UnitDebuff("target", spellName, nil, "PLAYER")
					info.auraEndTime = endTime
					info.auraDuration = duration
				end

				local cd = icon.cooldown
				local cdParent
				if (db.showCooldown and info.auraEndTime) then
					if (not cd) then
						cd = CreateFrame("Cooldown", nil, icon)
						cd:SetAllPoints()
						cd:SetReverse(true)
						icon.cooldown = cd
					else
						cd:Show()
					end

					cd:SetDrawEdge(db.drawEdge)
					cd:SetCooldown(info.auraEndTime - info.auraDuration, info.auraDuration)
					cdParent = cd

				elseif (cd) then
					cd:Hide()
					cdParent = icon
				end

				cd = icon.countdown
				if (db.showCountdown and info.auraEndTime) then
					if (not cd) then
						cd = cdParent:CreateFontString(nil, "OVERLAY", "GameFontNormal")
						cd:SetPoint("BOTTOM")
						cd:SetTextColor(1, 1, 1)
						icon.countdown = cd
					else
						cd:SetParent(cdParent)
						cd:Show()
					end

					icon.timeLeft = info.auraEndTime - GetTime()
					icon:SetScript("OnUpdate", countdownOnUpdate)

				elseif (cd) then
					cd:Hide()
					icon:SetScript("OnUpdate", nil)
				end
			end

			del(list)
		end

		for spellName,info in pairs(oldicons) do
			local icon = self.icons[spellName]
			self.icons[spellName] = nil
			assert(icon)
			self:FreeIcon(icon)
		end

		del(oldicons)
	end

	-- apbar.OnShow
	local function apbarOnShow(self)
		self:SetValue(module.currentAP)
	end

	-- apbar.OnHide
	local function apbarOnHide(self)
		for spellName,icon in pairs(self.icons) do
			self:FreeIcon(icon)
			self.icons[spellName] = nil
		end
	end

	-- CreateDisplay
	function module:CreateAPBar()
		local apbar = self:CreateBar("AP")

		apbar.icons = {}

		apbar:SetScript("OnHide", apbarOnHide)
		apbar:SetScript("OnShow", apbarOnShow)

		apbar.CreateIcon		= apbarCreateIcon
		apbar.AquireIcon		= apbarAquireIcon
		apbar.FreeIcon			= apbarFreeIcon

		apbar.SuperDraw			= apbar.Draw
		apbar.Draw				= apbarDraw

		self:RegisterCallback("ChangedAP", apbarEventChangedAP, apbar)
		self:RegisterCallback("AuraApplied", apbarEventAuraApplied, apbar)
		self:RegisterCallback("AuraRemoved", apbarEventAuraRemoved, apbar)

		return apbar
	end	
end

do
	-- frame.OnDragStart
	local function frameOnShow(self)
		self:RegisterEvent("MODIFIER_STATE_CHANGED")
		self.apBar:Show()
		self.apBar:SetSkin()
	end

	-- frame.OnDragStop
	local function frameOnHide(self)
		self:UnregisterAllEvents()
		self.apBar:Hide()
	end

	-- frame.OnDragStop
	local function frameOnDragStop(self)
		self:StopMovingOrSizing()
		module:SavePosition(self)
	end

	-- frameOnMouseWheel
	local function frameOnMouseWheel(self, delta)
		if (IsShiftKeyDown()) then
			local scale = self:GetScale()
			self:SetScale(scale + (delta / 10))
			db.scale = scale
		end
	end

	-- frame.OnEvent
	local function frameOnEvent(self, event, arg1, arg2)
		if (event == "MODIFIER_STATE_CHANGED") then
			if (arg1 == "LSHIFT" or arg1 == "RSHIFT") then
				if (arg2 == 1) then
					self:EnableMouse(true)
					self:EnableMouseWheel(true)
				else
					frameOnDragStop(self)
					self:EnableMouse(false)
					self:EnableMouseWheel(false)
				end
			end
		end
	end

	-- frame.OnUpdate
	local function frameOnUpdate(self, elapsed)
		-- Nasty to do this, but UNIT_STATS is only fired when stat modifiers are a given figure
		-- It is not for stats % modifiers (such as Unleashed Rage's 10%)
		-- Checking soon after UNIT_AURA is an option, but it seems you can't check directly AT UNIT_AURA
		module:DoStats()

		local a = self:GetAlpha()
		if (self.targetAlpha < a) then
			local a = self:GetAlpha() - elapsed * 2
			if (a <= self.targetAlpha) then
				a = self.targetAlpha
				if (a == 0) then
					self:Hide()
				end
			end
			self:SetAlpha(a)

		elseif (self.targetAlpha > a) then
			local a = self:GetAlpha() + elapsed * 2
			if (a >= self.targetAlpha) then
				a = self.targetAlpha
			end
			self:SetAlpha(a)
		end
	end

	local function frameOpen(self, targetAlpha)
		self.targetAlpha = targetAlpha
		if (not self:IsShown()) then
			self:Show()
			module:DoStats()
		end
	end

	local function frameClose(self)
		if (self:IsShown()) then
			self.targetAlpha = 0
		end
	end

	-- CreateDisplay
	function module:CreateDisplay()
		local f = CreateFrame("Frame", "BleedingHeartsFrame", UIParent)
		self.frame = f
		f:Hide()
		f:SetWidth(50)
		f:SetHeight(50)

		f:SetScale(db.scale or 1)

		f:SetMovable(true)
		f:RegisterForDrag("LeftButton")

		f:SetScript("OnShow",		frameOnShow)
		f:SetScript("OnHide",		frameOnHide)
		f:SetScript("OnDragStart",	f.StartMoving)
		f:SetScript("OnDragStop",	frameOnDragStop)
		f:SetScript("OnMouseWheel", frameOnMouseWheel)
		f:SetScript("OnEvent",		frameOnEvent)
		f:SetScript("OnUpdate",		frameOnUpdate)

		f.Open		= frameOpen
		f.Close		= frameClose

		f.apBar = self:CreateAPBar()

		module:RestorePosition(f)

		self.CreateDisplay = nil
	end
end

-- ShowFrame
function module:ShowFrame(targetAlpha)
	if (not self.frame) then
		self:CreateDisplay()
	end
	self.frame:SetScale(db.scale)
	self.frame:Open(targetAlpha)
end

-- HideFrame
function module:HideFrame()
	if (self.frame) then
		self.frame:Close()
	end
end

-- ShowWhen
function module:ShowWhen()
	local wantToShow
	if (db.show.always) then
		wantToShow = true
	else
		if (db.show.raid and GetNumRaidMembers() > 0) then
			wantToShow = true

		elseif (db.show.party and GetNumPartyMembers() > 0 and GetNumRaidMembers() == 0) then
			wantToShow = true

		elseif (db.show.solo and GetNumPartyMembers() == 0 and GetNumRaidMembers() == 0) then
			wantToShow = true
		end
	end

	if (wantToShow) then
		if (UnitAffectingCombat("player")) then
			wantToShow = db.inalpha
		else
			wantToShow = db.outalpha
		end
	end

	if (wantToShow and wantToShow > 0) then
		self:ShowFrame(wantToShow)
		return
	end

	self:HideFrame()
end

-- SavePosition
function module:SavePosition(frame)
	local point, _, relPoint, x, y = frame:GetPoint(1)
	if (point) then
		if (not self.db.char.pos) then
			self.db.char.pos = {}
		end
		self.db.char.pos[frame:GetName()] = {point, "UIParent", relPoint, x, y}
		if (frame:IsResizable()) then
			if (not self.db.char.frameSizes) then
				self.db.char.frameSizes = {}
			end
			self.db.char.frameSizes[frame:GetName()] = {frame:GetWidth(), frame:GetHeight()}
		end
	end
end

-- RestorePosition
function module:RestorePosition(frame)
	local pos = self.db.char.pos and self.db.char.pos[frame:GetName()]
	if (pos) then
		frame:SetPoint(unpack(pos))
		if (self.db.char.frameSizes) then
			if (frame:IsResizable()) then
				local size = self.db.char.frameSizes[frame:GetName()]
				if (size) then
					frame:SetWidth(size[1])
					frame:SetHeight(size[2])
				end
			else
				self.db.char.frameSizes[frame:GetName()] = nil
			end
		end
	else
		frame:SetPoint("CENTER")
	end
end

-- PLAYER_LEAVING_WORLD
function module:PLAYER_LEAVING_WORLD()
	self:ShowWhen()
end
module.PLAYER_ENTERING_WORLD	= module.PLAYER_LEAVING_WORLD
module.RAID_ROSTER_UPDATE		= module.PLAYER_LEAVING_WORLD
module.PARTY_MEMBERS_CHANGED	= module.PLAYER_LEAVING_WORLD

-- PLAYER_REGEN_ENABLED
function module:PLAYER_REGEN_ENABLED()
	self.active = deepDel(self.active)
	self.active = new()
	self:ShowWhen()
end

-- PLAYER_REGEN_DISABLED
function module:PLAYER_REGEN_DISABLED()
	self:ShowWhen()
end

-- DoStats
function module:DoStats()
	local base, posBuff, negBuff = UnitAttackPower("player")
	local AP = max(base + posBuff + negBuff, 0)

	if (AP ~= self.currentAP) then
		self.currentAP = AP
		self.callbacks:Fire("ChangedAP", AP)
	end
end

-- UNIT_STATS
function module:UNIT_STATS(e, unit)
	if (unit == "player") then
		self:DoStats()
	end
end

local clEvents = {}
-- SPELL_AURA_APPLIED
function clEvents:SPELL_AURA_APPLIED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags,
	spellId, spellName, spellSchool, auraType)

	if (sourceGUID == playerGUID) then
		if (not UnitInParty(destName) and not UnitInRaid(destName)) then
			local mon = self.monitor[spellName]
			if (mon) then
				local unitInfo = self.active[destGUID]
				if (not unitInfo) then
					unitInfo = new()
					self.active[destGUID] = unitInfo
				end

				local n = unitInfo[spellName]
				if (not n) then
					n = new()
					unitInfo[spellName] = n
				else
					if (event == "SPELL_AURA_REFRESH") then
						local now = GetTime()
						if (self.refreshAtOldAP and self.refreshAtOldAP >= now - 1.4) then
							-- Special refresh, keep existing AP value
							self.callbacks:Fire("AuraApplied", destGUID, spellName, spellId)
							return
						end
					end
				end

				n.when = timestamp
				n.ap = self.currentAP
				n.id = spellId
				n.auraEndTime = nil
				n.auraDuration = nil
				if (mon == "ROGUE") then
					local name, rank = GetSpellInfo(spellId)
					rank = tonumber(rank:match("(%d+)"))
					assert(rank)

					n.combos = self.combos
					n.value = RuptureCalcs(rank, n.ap, n.combos)
					self.lastRuptureRank = rank
					self.combos = nil
				end

				self.callbacks:Fire("AuraApplied", destGUID, spellName, spellId)
			end
		end
	end
end
clEvents.SPELL_AURA_REFRESH = clEvents.SPELL_AURA_APPLIED

-- SPELL_AURA_REMOVED
function clEvents:SPELL_AURA_REMOVED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags,
	spellId, spellName, spellSchool, auraType)

	if (sourceGUID == playerGUID) then
		if (not UnitInParty(destName) and not UnitInRaid(destName)) then
			local unitInfo = self.active[destGUID]
			if (unitInfo) then
				local active = unitInfo[spellName]
				if (active) then
					self.callbacks:Fire("AuraRemoved", destGUID, spellName, spellId)
					unitInfo[spellName] = del(active)
				end
			end
		end
	end
end

-- COMBAT_LOG_EVENT_UNFILTERED
function module:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	local f = clEvents[event]
	if (f) then
		f(self, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	end
end

-- HasGlyph
local function HasGlyph(id)
	local group = GetActiveTalentGroup()
	for i = 1,GetNumGlyphSockets() do
		local enabled, glyphType, glyphSpell, icon = GetGlyphSocketInfo(i, group)
		if (glyphSpell == id) then
			return true
		end
	end
end

local rupture = GetSpellInfo(1943)
local pestilence = GetSpellInfo(50842)
local scourgeStrike = GetSpellInfo(55090)

-- UNIT_SPELLCAST_SUCCEEDED
function module:UNIT_SPELLCAST_SUCCEEDED(e, unit, spell)
	if (unit == "player") then
		if (playerClass == "ROGUE") then
			if (spell == rupture) then
				self.combos = GetComboPoints("player")
			end
		elseif (playerClass == "DEATHKNIGHT") then
			if (spell == pestilence) then
				if (HasGlyph(63334)) then			-- Glyph of Disease
					-- Refreshes the active dots instead of re-applying at current AP level
					self.refreshAtOldAP = GetTime()
				end
			elseif (spell == scourgeStrike) then
				if (HasGlyph(58642)) then			-- Glyph of Scourge Strike
					-- Extends the time on the existing dots instead of re-applying at current AP level
					self.refreshAtOldAP = GetTime()
				end
			end
		end
	end
end

-- UNIT_COMBO_POINTS
function module:UNIT_COMBO_POINTS(e, unit)
	if (unit == "player") then
		if (self.frame and self.frame:IsShown()) then
			self.frame.apBar:Refresh()
		end
	end
end

-- PLAYER_TARGET_CHANGED
function module:PLAYER_TARGET_CHANGED()
	if (self.frame and self.frame:IsShown()) then
		self.frame.apBar:Refresh()
	end
end

-- ChangeAlphaOption
function module:ChangeAlphaOption(info, val)
	db.alpha = max(0.1, min(1, val))
	if (self.frame) then
		self.frame:SetAlpha(db.alpha)
	end
end

-- OpenConfig
function module:OpenConfig(input)
	InterfaceOptionsFrameAddOns.scrollFrame:SetVerticalScroll(0)
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.BleedingHearts)
end

-- SetupOptions
function module:SetupOptions()
	self.optionsFrames = {}

	-- setup options table
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("BleedingHearts", self.options)
	local ACD3 = LibStub("AceConfigDialog-3.0")

	self.optionsFrames.BleedingHearts = ACD3:AddToBlizOptions("BleedingHearts", nil, nil, "General")
	self:RegisterChatCommand("bleedinghearts", "OpenConfig")

	self:RegisterModuleOptions("Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db), L["Profiles"])
	-- Add ordering data to the option table generated by AceDBOptions-3.0
	self.options.args.Profiles.order = -2

	self.SetupOptions = nil
end

-- RegisterModuleOptions
function module:RegisterModuleOptions(name, optionTbl, displayName)
	self.options.args[name] = (type(optionTbl) == "function") and optionTbl() or optionTbl
	self.optionsFrames[name] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("BleedingHearts", displayName, "BleedingHearts", name)
end

-- OnProfileChanged
function module:OnProfileChanged(event, database, newProfileKey)
	db = database.profile
end

-- OnModuleInitialize
function module:OnInitialize()
	playerClass = select(2, UnitClass("player"))
	playerName = UnitName("player")
	playerGUID = UnitGUID("player")

	self.db = LibStub("AceDB-3.0"):New("BleedingHeartsDB", defaults, "Default")
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")

	self.db:RegisterDefaults({
		profile = {
			scale = 1,
			inalpha = 0.8,
			outalpha = 0,
			showCooldown = true,
			drawEdge = true,
			showCountdown = true,
			minCountdown = 10,
			show = {
				always = false,
				combat = true,
				solo = false,
				party = false,
				raid = true,
			},
			skin = "thermometer",
			flattexture = "Blizzard",
			flatwidth = 250,
		}
	})
	db = self.db.profile

	self.callbacks = LibStub("CallbackHandler-1.0"):New(self)

	self.OnInitialize = nil
end

-- OnModuleEnable
function module:OnEnable()
	self.active = {}
	if (self.db) then
		local class = select(2, UnitClass("player"))
		if (class ~= playerClass and self.OnModuleInitialize) then
			self:OnModuleInitialize()
		end

		playerClass = select(2, UnitClass("player"))
		playerName = UnitName("player")
		playerGUID = UnitGUID("player")

		self:SetupOptions()

		self:RegisterEvent("PLAYER_LEAVING_WORLD")
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("RAID_ROSTER_UPDATE")
		self:RegisterEvent("PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("UNIT_STATS")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")

		if (playerClass == "ROGUE") then
			self:RegisterEvent("UNIT_COMBO_POINTS")
		end

		self:DoStats()
		self:ShowWhen()

		self.title = L["TITLE"].." r"..module.version
	end
end

-- OnDisable
function module:OnDisable()
	self.active = nil
end
